@@ -1,16 +1,18 @@
 """Create regression estimators based on the sparse group lasso."""
+import logging
 import numpy as np
 
 from functools import partial
-from joblib import delayed, effective_n_jobs
+from joblib import delayed, effective_n_jobs, Parallel
 from scipy import sparse
 from scipy.optimize import root_scalar
+from skopt import BayesSearchCV
 from tqdm.auto import tqdm
 
 from sklearn.base import RegressorMixin, TransformerMixin
 from sklearn.linear_model._base import LinearModel, _preprocess_data
 from sklearn.linear_model._coordinate_descent import _alpha_grid as _lasso_alpha_grid
-from sklearn.linear_model._coordinate_descent import _path_residuals
+from sklearn.metrics import get_scorer
 from sklearn.model_selection import check_cv
 from sklearn.utils.extmath import safe_sparse_dot
 from sklearn.utils.fixes import _joblib_parallel_args
@@ -18,9 +20,10 @@ from sklearn.utils.validation import check_array, check_is_fitted, column_or_1d
 
 from ._base import SGLBaseEstimator
 from ._prox import _soft_threshold
-from .utils import check_groups, _ProgressParallel
+from .utils import check_groups
 
 __all__ = ["SGL", "sgl_path", "SGLCV"]
+logger = logging.getLogger(__name__)
 
 
 class SGL(SGLBaseEstimator, RegressorMixin, LinearModel):
@@ -171,7 +174,9 @@ def _alpha_grid(
         Target values
 
     Xy : array-like of shape (n_features,), default=None
-        Xy = np.dot(X.T, y) that can be precomputed.
+        Xy = np.dot(X.T, y) that can be precomputed. If supplying ``Xy``,
+        prevent train/test leakage by ensuring the ``Xy`` is precomputed
+        using only training data.
 
     groups : list of numpy.ndarray
         list of arrays of non-overlapping indices for each group. For
@@ -291,7 +296,7 @@ def _alpha_grid(
     alpha_max = np.max(min_alphas) * 1.2
 
     # Test feature sparsity just to make sure we're on the right side of the root
-    while (
+    while (  # pragma: no cover
         model(
             groups=groups,
             alpha=alpha_max,
@@ -303,7 +308,7 @@ def _alpha_grid(
         .chosen_features_.size
         > 0
     ):
-        alpha_max *= 1.2
+        alpha_max *= 1.2  # pragma: no cover
 
     if alpha_max <= np.finfo(float).resolution:
         alphas = np.empty(n_alphas)
@@ -380,7 +385,9 @@ def sgl_path(
         If None alphas are set automatically.
 
     Xy : array-like of shape (n_features,), default=None
-        Xy = np.dot(X.T, y) that can be precomputed.
+        Xy = np.dot(X.T, y) that can be precomputed. If supplying ``Xy``,
+        prevent train/test leakage by ensuring the ``Xy`` is precomputed
+        using only training data.
 
     normalize : bool, default=False
         This parameter is ignored when ``fit_intercept`` is set to False.
@@ -396,9 +403,6 @@ def sgl_path(
     verbose : bool or int, default=False
         Amount of verbosity.
 
-    return_n_iter : bool, default=False
-        Whether to return the number of iterations or not.
-
     check_input : bool, default=True
         Skip input validation checks, assuming there are handled by the
         caller when check_input=False.
@@ -408,16 +412,16 @@ def sgl_path(
 
     Returns
     -------
+    coefs : ndarray of shape (n_features, n_alphas) or (n_features + 1, n_alphas)
+        List of coefficients for the Logistic Regression model. If
+        fit_intercept is set to True then the first dimension will be
+        n_features + 1, where the last item represents the intercept.
+
     alphas : ndarray of shape (n_alphas,)
         The alphas along the path where models are computed.
 
-    coefs : ndarray of shape (n_features, n_alphas)
-        Coefficients along the path.
-
-    n_iters : list of int
-        The number of iterations taken by the PGD solver to
-        reach the specified tolerance for each alpha.
-        (Is returned when ``return_n_iter`` is set to True).
+    n_iters : array of shape (n_alphas,)
+        Actual number of iteration for each alpha.
 
     See Also
     --------
@@ -475,7 +479,11 @@ def sgl_path(
     max_iter = params.get("max_iter", 1000)
     loss = params.get("loss", "squared_loss")
     n_iters = np.empty((n_alphas,), dtype=int)
-    coefs = np.empty((n_features, n_alphas), dtype=X.dtype)
+
+    if fit_intercept:
+        coefs = np.empty((n_features + 1, n_alphas), dtype=X.dtype)
+    else:
+        coefs = np.empty((n_features, n_alphas), dtype=X.dtype)
 
     model = SGL(
         l1_ratio=l1_ratio,
@@ -500,7 +508,11 @@ def sgl_path(
         model.set_params(alpha=alpha)
         model.fit(X, y, loss=loss)
 
-        coefs[..., i] = model.coef_
+        if fit_intercept:
+            coefs[..., i] = np.concatenate([model.coef_, [model.intercept_]])
+        else:
+            coefs[..., i] = model.coef_
+
         n_iters[i] = model.n_iter_
 
         if verbose:
@@ -509,13 +521,189 @@ def sgl_path(
             elif verbose > 1:
                 print("Path: %03i out of %03i" % (i, n_alphas))
 
-    # TODO: Compute dual gaps here
-    dual_gaps = None
+    return coefs, alphas, n_iters
 
-    if return_n_iter:
-        return alphas, coefs, dual_gaps, n_iters
 
-    return alphas, coefs, dual_gaps
+def sgl_scoring_path(
+    X,
+    y,
+    train,
+    test,
+    l1_ratio=0.5,
+    groups=None,
+    scale_l2_by="group_length",
+    eps=1e-3,
+    n_alphas=100,
+    alphas=None,
+    Xy=None,
+    normalize=False,
+    copy_X=True,
+    verbose=False,
+    check_input=True,
+    scoring=None,
+    **params,
+):
+    """Return the scores for the models computed by 'path'.
+
+    Parameters
+    ----------
+    X : {array-like, sparse matrix} of shape (n_samples, n_features)
+        Training data.
+
+    y : array-like of shape (n_samples,) or (n_samples, n_targets)
+        Target values.
+
+    train : list of indices
+        The indices of the train set.
+
+    test : list of indices
+        The indices of the test set.
+
+    l1_ratio : float, default=0.5
+        Number between 0 and 1 passed to SGL estimator (scaling between the
+        group lasso and lasso penalties). ``l1_ratio=1`` corresponds to the
+        Lasso.
+
+    groups : list of numpy.ndarray
+        list of arrays of non-overlapping indices for each group. For
+        example, if nine features are grouped into equal contiguous groups of
+        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
+        array([6, 7, 8])]``. If the feature matrix contains a bias or
+        intercept feature, do not include it as a group. If None, all
+        features will belong to one group.
+
+    scale_l2_by : ["group_length", None], default="group_length"
+        Scaling technique for the group-wise L2 penalty.
+        By default, ``scale_l2_by="group_length`` and the L2 penalty is
+        scaled by the square root of the group length so that each variable
+        has the same effect on the penalty. This may not be appropriate for
+        one-hot encoded features and ``scale_l2_by=None`` would be more
+        appropriate for that case. ``scale_l2_by=None`` will also reproduce
+        ElasticNet results when all features belong to one group.
+
+    eps : float, default=1e-3
+        Length of the path. ``eps=1e-3`` means that
+        ``alpha_min / alpha_max = 1e-3``.
+
+    n_alphas : int, default=100
+        Number of alphas along the regularization path.
+
+    alphas : ndarray, default=None
+        List of alphas where to compute the models.
+        If None alphas are set automatically.
+
+    Xy : array-like of shape (n_features,), default=None
+        Xy = np.dot(X.T, y) that can be precomputed. If supplying ``Xy``,
+        prevent train/test leakage by ensuring the ``Xy`` is precomputed
+        using only training data.
+
+    normalize : bool, default=False
+        This parameter is ignored when ``fit_intercept`` is set to False.
+        If True, the regressors X will be normalized before regression by
+        subtracting the mean and dividing by the l2-norm.
+        If you wish to standardize, please use
+        :class:`sklearn.preprocessing.StandardScaler` before calling ``fit``
+        on an estimator with ``normalize=False``.
+
+    copy_X : bool, default=True
+        If ``True``, X will be copied; else, it may be overwritten.
+
+    verbose : bool or int, default=False
+        Amount of verbosity.
+
+    check_input : bool, default=True
+        Skip input validation checks, assuming there are handled by the
+        caller when check_input=False.
+
+    scoring : callable, default=None
+        A string (see sklearn model evaluation documentation) or a scorer
+        callable object / function with signature ``scorer(estimator, X, y)``.
+        For a list of scoring functions that can be used, look at
+        `sklearn.metrics`. The default scoring option used is accuracy_score.
+
+    **params : kwargs
+        Keyword arguments passed to the SGL estimator
+
+    Returns
+    -------
+    coefs : ndarray of shape (n_features, n_alphas) or (n_features + 1, n_alphas)
+        List of coefficients for the SGL model. If fit_intercept is set to
+        True then the first dimension will be n_features + 1, where the last
+        item represents the intercept.
+
+    alphas : ndarray
+        Grid of alphas used for cross-validation.
+
+    scores : ndarray of shape (n_alphas,)
+        Scores obtained for each alpha.
+
+    n_iter : ndarray of shape(n_alphas,)
+        Actual number of iteration for each alpha.
+    """
+    if Xy is not None:
+        logger.warning(
+            "You supplied the `Xy` parameter. Remember to ensure "
+            "that Xy is computed from the training data alone."
+        )
+
+    X_train = X[train]
+    y_train = y[train]
+    X_test = X[test]
+    y_test = y[test]
+
+    coefs, alphas, n_iter = sgl_path(
+        X_train,
+        y_train,
+        l1_ratio=l1_ratio,
+        groups=groups,
+        scale_l2_by=scale_l2_by,
+        eps=eps,
+        n_alphas=n_alphas,
+        alphas=alphas,
+        Xy=Xy,
+        normalize=normalize,
+        copy_X=copy_X,
+        verbose=verbose,
+        check_input=False,
+        **params,
+    )
+    del X_train, y_train
+
+    fit_intercept = params.get("fit_intercept", True)
+    max_iter = params.get("max_iter", 1000)
+    tol = params.get("tol", 1e-7)
+
+    model = SGL(
+        l1_ratio=l1_ratio,
+        groups=groups,
+        scale_l2_by=scale_l2_by,
+        fit_intercept=fit_intercept,
+        max_iter=max_iter,
+        tol=tol,
+        warm_start=True,
+        verbose=False,
+        suppress_solver_warnings=True,
+        include_solver_trace=False,
+    )
+
+    if scoring is None:
+        scoring = "neg_mean_squared_error"
+
+    model.is_fitted_ = True
+
+    scores = list()
+    scoring = get_scorer(scoring)
+    for w in coefs.T:
+        if fit_intercept:
+            model.coef_ = w[:-1]
+            model.intercept_ = w[-1]
+        else:
+            model.coef_ = w
+            model.intercept_ = 0.0
+
+        scores.append(scoring(model, X_test, y_test))
+
+    return coefs, alphas, np.array(scores), n_iter
 
 
 class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
@@ -614,6 +802,37 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         ``None`` means 1 unless in a :obj:`joblib:joblib.parallel_backend` context.
         ``-1`` means using all processors.
 
+    tuning_strategy : ["grid", "bayes"], default="grid"
+        Hyperparameter tuning strategy to use. If ``tuning_strategy == "grid"``,
+        then evaluate all parameter points on the ``l1_ratio`` and ``alphas`` grid,
+        using warm start to evaluate different ``alpha`` values along the
+        regularization path. If ``tuning_strategy == "bayes"``, then a fixed
+        number of parameter settings is sampled using ``skopt.BayesSearchCV``.
+        The fixed number of settings is set by ``n_bayes_iter``. The
+        ``l1_ratio`` setting is sampled uniformly from the minimum and maximum
+        of the input ``l1_ratio`` parameter. The ``alpha`` setting is sampled
+        log-uniformly either from the maximum and minumum of the input ``alphas``
+        parameter, if provided or from ``eps`` * max_alpha to max_alpha where
+        max_alpha is a conservative estimate of the maximum alpha for which the
+        solution coefficients are non-trivial.
+
+    n_bayes_iter : int, default=50
+        Number of parameter settings that are sampled if using Bayes search
+        for hyperparameter optimization. ``n_bayes_iter`` trades off runtime
+        vs quality of the solution. Consider increasing ``n_bayes_points`` if
+        you want to try more parameter settings in parallel.
+
+    n_bayes_points : int, default=1
+        Number of parameter settings to sample in parallel if using Bayes
+        search for hyperparameter optimization. If this does not align with
+        ``n_bayes_iter``, the last iteration will sample fewer points.
+
+    random_state : int, RandomState instance or None, optional (default=None)
+        If int, random_state is the seed used by the random number generator;
+        If RandomState instance, random_state is the random number generator;
+        If None, the random number generator is the RandomState instance used
+        by `np.random`.
+
     Attributes
     ----------
     alpha_ : float
@@ -629,7 +848,7 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
     intercept_ : float or ndarray of shape (n_targets, n_features)
         Independent term in the decision function.
 
-    mse_path_ : ndarray of shape (n_l1_ratio, n_alpha, n_folds)
+    scoring_path_ : ndarray of shape (n_l1_ratio, n_alpha, n_folds)
         Mean square error for the test set on each fold, varying l1_ratio and
         alpha.
 
@@ -640,6 +859,11 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         number of iterations run by the proximal gradient descent solver to
         reach the specified tolerance for the optimal alpha.
 
+    bayes_optimizer_ : skopt.BayesSearchCV instance or None
+        The BayesSearchCV instance used for hyperparameter optimization if
+        ``tuning_strategy == "bayes"``. If ``tuning_strategy == "grid"``,
+        then this attribute is None.
+
     See Also
     --------
     sgl_path
@@ -662,6 +886,10 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         cv=None,
         verbose=False,
         n_jobs=None,
+        tuning_strategy="grid",
+        n_bayes_iter=50,
+        n_bayes_points=1,
+        random_state=None,
     ):
         self.l1_ratio = l1_ratio
         self.groups = groups
@@ -677,6 +905,10 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         self.cv = cv
         self.verbose = verbose
         self.n_jobs = n_jobs
+        self.tuning_strategy = tuning_strategy
+        self.n_bayes_iter = n_bayes_iter
+        self.n_bayes_points = n_bayes_points
+        self.random_state = random_state
 
     def fit(self, X, y):
         """Fit sparse group lasso linear model.
@@ -693,6 +925,12 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         y : array-like of shape (n_samples,) or (n_samples, n_targets)
             Target values
         """
+        if self.tuning_strategy.lower() not in ["grid", "bayes"]:
+            raise ValueError(
+                "`tuning_strategy` must be either 'grid' or 'bayes'; got "
+                "{0} instead.".format(self.tuning_strategy)
+            )
+
         # This makes sure that there is no duplication in memory.
         # Dealing right with copy_X is important in the following:
         # Multiple functions touch X and subsamples of X and can induce a
@@ -755,6 +993,10 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         path_params = self.get_params()
         path_params.pop("cv", None)
         path_params.pop("n_jobs", None)
+        path_params.pop("tuning_strategy", None)
+        path_params.pop("n_bayes_iter", None)
+        path_params.pop("n_bayes_points", None)
+        path_params.pop("random_state", None)
 
         l1_ratios = np.atleast_1d(path_params["l1_ratio"])
         # For the first path, we need to set l1_ratio
@@ -792,95 +1034,150 @@ class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
         if effective_n_jobs(self.n_jobs) > 1:
             path_params["copy_X"] = False
 
-        # "precompute" has no effect but it is expected by _path_residuals
-        path_params["precompute"] = False
-
         if isinstance(self.verbose, int):  # pragma: no cover
             path_params["verbose"] = self.verbose - 1
 
         # init cross-validation generator
         cv = check_cv(self.cv)
 
-        # Compute path for all folds and compute MSE to get the best alpha
-        folds = list(cv.split(X, y))
-        best_mse = np.inf
+        if self.tuning_strategy == "grid":
+            # Compute path for all folds and compute MSE to get the best alpha
+            folds = list(cv.split(X, y))
+            best_score = -np.inf
 
-        # We do a double for loop folded in one, in order to be able to
-        # iterate in parallel on l1_ratio and folds
-        jobs = (
-            delayed(_path_residuals)(
-                X,
-                y,
-                train,
-                test,
-                sgl_path,
-                path_params,
-                alphas=this_alphas,
-                l1_ratio=this_l1_ratio,
-                X_order="F",
-                dtype=X.dtype.type,
+            path_params.pop("cv", None)
+            path_params.pop("n_jobs", None)
+            path_params.pop("alphas", None)
+            path_params.pop("l1_ratio", None)
+            path_params.update({"groups": groups})
+
+            # We do a double for loop folded in one, in order to be able to
+            # iterate in parallel on l1_ratio and folds
+            jobs = (
+                delayed(sgl_scoring_path)(
+                    X=X,
+                    y=y,
+                    train=train,
+                    test=test,
+                    l1_ratio=this_l1_ratio,
+                    alphas=this_alphas,
+                    Xy=None,
+                    **path_params,
+                )
+                for this_l1_ratio, this_alphas in zip(l1_ratios, alphas)
+                for train, test in folds
             )
-            for this_l1_ratio, this_alphas in zip(l1_ratios, alphas)
-            for train, test in folds
-        )
 
-        if isinstance(self.verbose, int):  # pragma: no cover
-            parallel_verbosity = self.verbose - 2
-            if parallel_verbosity < 0:
-                parallel_verbosity = 0
-        else:  # pragma: no cover
-            parallel_verbosity = self.verbose
+            if isinstance(self.verbose, int):  # pragma: no cover
+                parallel_verbosity = self.verbose - 2
+                if parallel_verbosity < 0:
+                    parallel_verbosity = 0
+            else:  # pragma: no cover
+                parallel_verbosity = self.verbose
 
-        mse_paths = _ProgressParallel(
-            n_jobs=self.n_jobs,
-            verbose=parallel_verbosity,
-            use_tqdm=bool(self.verbose),
-            desc="L1 ratios * CV folds",
-            total=n_l1_ratio * len(folds),
-            **_joblib_parallel_args(prefer="threads"),
-        )(jobs)
+            score_paths = Parallel(
+                n_jobs=self.n_jobs,
+                verbose=parallel_verbosity,
+                **_joblib_parallel_args(prefer="threads"),
+            )(jobs)
 
-        mse_paths = np.reshape(mse_paths, (n_l1_ratio, len(folds), -1))
-        mean_mse = np.mean(mse_paths, axis=1)
-        self.mse_path_ = np.squeeze(np.rollaxis(mse_paths, 2, 1))
+            coefs_paths, alphas_paths, scores, n_iters = zip(*score_paths)
+            scores = np.reshape(scores, (n_l1_ratio, len(folds), -1))
+            alphas_paths = np.reshape(alphas_paths, (n_l1_ratio, len(folds), -1))
+            n_iters = np.reshape(n_iters, (n_l1_ratio, len(folds), -1))
+            coefs_paths = np.reshape(
+                coefs_paths, (n_l1_ratio, len(folds), -1, n_alphas)
+            )
 
-        for l1_ratio, l1_alphas, mse_alphas in zip(l1_ratios, alphas, mean_mse):
-            i_best_alpha = np.argmin(mse_alphas)
-            this_best_mse = mse_alphas[i_best_alpha]
-            if this_best_mse < best_mse:
-                best_alpha = l1_alphas[i_best_alpha]
-                best_l1_ratio = l1_ratio
-                best_mse = this_best_mse
+            mean_score = np.mean(scores, axis=1)
+            self.scoring_path_ = np.squeeze(np.moveaxis(scores, 2, 1))
 
-        self.l1_ratio_ = best_l1_ratio
-        self.alpha_ = best_alpha
+            for l1_ratio, l1_alphas, score_alphas in zip(l1_ratios, alphas, mean_score):
+                i_best_alpha = np.argmax(score_alphas)
+                this_best_score = score_alphas[i_best_alpha]
+                if this_best_score > best_score:
+                    best_alpha = l1_alphas[i_best_alpha]
+                    best_l1_ratio = l1_ratio
+                    best_score = this_best_score
 
-        if self.alphas is None:
-            self.alphas_ = np.asarray(alphas)
-            if n_l1_ratio == 1:
-                self.alphas_ = self.alphas_[0]
-        # Remove duplicate alphas in case alphas is provided.
+            self.l1_ratio_ = best_l1_ratio
+            self.alpha_ = best_alpha
+
+            if self.alphas is None:
+                self.alphas_ = np.asarray(alphas)
+                if n_l1_ratio == 1:
+                    self.alphas_ = self.alphas_[0]
+            # Remove duplicate alphas in case alphas is provided.
+            else:
+                self.alphas_ = np.asarray(alphas[0])
+
+            # Refit the model with the parameters selected
+            common_params = {
+                name: value
+                for name, value in self.get_params().items()
+                if name in model.get_params()
+            }
+
+            model.set_params(**common_params)
+            model.alpha = best_alpha
+            model.l1_ratio = best_l1_ratio
+            model.copy_X = copy_X
+
+            model.fit(X, y)
+
+            self.coef_ = model.coef_
+            self.intercept_ = model.intercept_
+            self.n_iter_ = model.n_iter_
+            self.bayes_optimizer_ = None
+            self.is_fitted_ = True
         else:
-            self.alphas_ = np.asarray(alphas[0])
+            # Set the model with the common input params
+            common_params = {
+                name: value
+                for name, value in self.get_params().items()
+                if name in model.get_params()
+            }
+            model.set_params(**common_params)
+            model.copy_X = copy_X
 
-        # Refit the model with the parameters selected
-        common_params = {
-            name: value
-            for name, value in self.get_params().items()
-            if name in model.get_params()
-        }
+            search_spaces = {"alpha": (np.min(alphas), np.max(alphas), "log-uniform")}
 
-        model.set_params(**common_params)
-        model.alpha = best_alpha
-        model.l1_ratio = best_l1_ratio
-        model.copy_X = copy_X
+            l1_min = np.min(self.l1_ratio)
+            l1_max = np.max(self.l1_ratio)
 
-        model.fit(X, y)
+            if l1_min < l1_max:
+                search_spaces["l1_ratio"] = (
+                    np.min(self.l1_ratio),
+                    np.max(self.l1_ratio),
+                    "uniform",
+                )
+
+            self.bayes_optimizer_ = BayesSearchCV(
+                model,
+                search_spaces,
+                n_iter=self.n_bayes_iter,
+                cv=cv,
+                n_jobs=self.n_jobs,
+                n_points=self.n_bayes_points,
+                verbose=self.verbose,
+                random_state=self.random_state,
+                return_train_score=True,
+                scoring="neg_mean_squared_error",
+                error_score=-np.inf,
+            )
+
+            self.bayes_optimizer_.fit(X, y)
+
+            self.alpha_ = self.bayes_optimizer_.best_estimator_.alpha
+            self.l1_ratio_ = self.bayes_optimizer_.best_estimator_.l1_ratio
+            self.coef_ = self.bayes_optimizer_.best_estimator_.coef_
+            self.intercept_ = self.bayes_optimizer_.best_estimator_.intercept_
+            self.n_iter_ = self.bayes_optimizer_.best_estimator_.n_iter_
+            self.is_fitted_ = True
+            self.scoring_path_ = None
+            param_alpha = self.bayes_optimizer_.cv_results_["param__alpha"]
+            self.alphas_ = np.sort(param_alpha)[::-1]
 
-        self.coef_ = model.coef_
-        self.intercept_ = model.intercept_
-        self.n_iter_ = model.n_iter_
-        self.is_fitted_ = True
         return self
 
     @property
