@@ -1,11 +1,13 @@
 import numpy as np
 import pytest
 
-from groupyr import SGL, SGLCV, LogisticSGLCV
-from groupyr.datasets import make_group_regression
+from groupyr import SGL, SGLCV, LogisticSGL, LogisticSGLCV
+from groupyr.datasets import make_group_regression, make_group_classification
 from groupyr.sgl import _alpha_grid
 
+from skopt import BayesSearchCV
 from sklearn.linear_model.tests.test_coordinate_descent import build_dataset
+from sklearn.model_selection import GridSearchCV, StratifiedKFold
 from sklearn.utils._testing import assert_almost_equal, assert_array_almost_equal
 
 
@@ -29,7 +31,7 @@ def test_sgl_input_validation():
         _alpha_grid(X, y, l1_ratio=0.5, scale_l2_by="error")
 
 
-@pytest.mark.parametrize("Estimator", [SGL, SGLCV, LogisticSGLCV])
+@pytest.mark.parametrize("Estimator", [SGL, SGLCV, LogisticSGL, LogisticSGLCV])
 def test_sgl_masks(Estimator):
     groups = [np.arange(5), np.arange(5, 10)]
     model = Estimator(groups=groups)
@@ -133,17 +135,50 @@ def test_warm_start(fit_intercept):
     assert_array_almost_equal(fx_warm_start, fx_cold_start)
 
 
-def test_sgl_cv():
+@pytest.mark.parametrize("tuning_strategy", ["grid", "bayes"])
+def test_sgl_cv(tuning_strategy):
     X, y, X_test, y_test = build_dataset()
-    max_iter = 150
-    clf = SGLCV(n_alphas=10, eps=1e-3, max_iter=max_iter, cv=3).fit(X, y)
+    max_iter = 500
+    clf = SGLCV(
+        n_alphas=10,
+        eps=1e-3,
+        max_iter=max_iter,
+        cv=3,
+        tuning_strategy=tuning_strategy,
+        random_state=42,
+        n_bayes_iter=10,
+    ).fit(X, y)
     assert_almost_equal(clf.alpha_, 0.056, 2)
     assert clf.score(X_test, y_test) > 0.99  # nosec
 
-    alphas = np.copy(clf.alphas_)
-    np.random.default_rng().shuffle(alphas)
-    clf2 = SGLCV(alphas=alphas, max_iter=max_iter, cv=3, n_jobs=2).fit(X, y)
-    assert_array_almost_equal(clf.alphas_, clf2.alphas_)
+    if tuning_strategy == "grid":
+        alphas = np.copy(clf.alphas_)
+        np.random.default_rng().shuffle(alphas)
+        clf2 = SGLCV(
+            alphas=alphas,
+            max_iter=max_iter,
+            cv=3,
+            n_jobs=2,
+            tuning_strategy=tuning_strategy,
+        ).fit(X, y)
+        assert_array_almost_equal(clf.alphas_, clf2.alphas_)
+
+        # Make sure that this attr is set to None:
+        assert clf2.bayes_optimizer_ is None
+
+    else:
+        clf = SGLCV(
+            l1_ratio=[0.95, 1.0],
+            n_alphas=10,
+            eps=1e-3,
+            max_iter=max_iter,
+            cv=3,
+            tuning_strategy=tuning_strategy,
+            random_state=42,
+            n_bayes_iter=20,
+        ).fit(X, y)
+        assert clf.score(X_test, y_test) > 0.98  # nosec
+        assert_almost_equal(clf.alpha_, 0.06, 2)
 
 
 @pytest.mark.parametrize("execution_number", range(5))
@@ -178,3 +213,61 @@ def test_sglcv_value_errors():
 
     with pytest.raises(ValueError):
         SGLCV().fit(X, y + [0])
+
+    with pytest.raises(ValueError):
+        SGLCV(tuning_strategy="error").fit(X, y)
+
+    with pytest.raises(ValueError):
+        LogisticSGLCV(tuning_strategy="error").fit(X, y)
+
+
+def test_LogisticSGLCV_GridSearchCV():
+    # make sure LogisticSGLCV gives same best params (l1_ratio and alpha) as
+    # GridSearchCV
+    X, y, groups = make_group_classification(random_state=42)
+
+    cv = StratifiedKFold(3)
+    l1_ratios = np.linspace(0, 1, 3)
+    alphas = np.logspace(-4, 4, 3)
+
+    clf_cv = LogisticSGLCV(alphas=alphas, l1_ratio=l1_ratios, groups=groups, cv=cv)
+    clf_cv.fit(X, y)
+
+    param_grid = {"alpha": alphas, "l1_ratio": l1_ratios}
+    clf = LogisticSGL(groups=groups)
+    gs = GridSearchCV(clf, param_grid, cv=cv)
+    gs.fit(X, y)
+
+    assert gs.best_params_["l1_ratio"] == clf_cv.l1_ratio_
+    assert gs.best_params_["alpha"] == clf_cv.alpha_
+
+
+def test_LogisticSGLCV_BayesSearchCV():
+    # make sure LogisticSGLCV gives same best params (l1_ratio and alpha) as
+    # BayesSearchCV
+    X, y, groups = make_group_classification(random_state=42)
+    cv = StratifiedKFold(3)
+
+    l1_ratios = np.linspace(0, 1, 3)
+    alphas = np.logspace(-4, 4, 3)
+    clf_cv = LogisticSGLCV(
+        alphas=alphas,
+        l1_ratio=l1_ratios,
+        groups=groups,
+        cv=cv,
+        tuning_strategy="bayes",
+        n_bayes_iter=10,
+        random_state=42,
+    )
+    clf_cv.fit(X, y)
+
+    search_spaces = {
+        "alpha": (np.min(alphas), np.max(alphas), "log-uniform"),
+        "l1_ratio": (np.min(l1_ratios), np.max(l1_ratios), "uniform"),
+    }
+    clf = LogisticSGL(groups=groups)
+    gs = BayesSearchCV(clf, search_spaces, cv=cv, random_state=42, n_iter=10)
+    gs.fit(X, y)
+
+    assert gs.best_params_["l1_ratio"] == clf_cv.l1_ratio_
+    assert gs.best_params_["alpha"] == clf_cv.alpha_
